// Based on dune/tnnmg/problem-classes/blocknonlineargsproblem.hh

#ifndef MY_BLOCK_PROBLEM_HH
#define MY_BLOCK_PROBLEM_HH

#include <dune/common/bitsetvector.hh>
#include <dune/common/parametertree.hh>

#include <dune/tnnmg/problem-classes/bisection.hh>
#include <dune/tnnmg/problem-classes/onedconvexfunction.hh>

#include "localnonlinearity.hh"
#include "samplefunctional.hh"

/** \brief Base class for problems where each block can be solved with a
 * modified gradient method */
template <class MyConvexProblemTypeTEMPLATE> class MyBlockProblem {
public:
  typedef MyConvexProblemTypeTEMPLATE MyConvexProblemType;
  typedef typename MyConvexProblemType::VectorType VectorType;
  typedef typename MyConvexProblemType::MatrixType MatrixType;
  typedef typename MyConvexProblemType::LocalVectorType LocalVectorType;
  typedef typename MyConvexProblemType::LocalMatrixType LocalMatrixType;

  int static const block_size = MyConvexProblemType::block_size;

  /** \brief Solves one local system using a modified gradient method */
  class IterateObject;

  MyBlockProblem(Dune::ParameterTree const &parset,
                 MyConvexProblemType &problem)
      : parset(parset), problem(problem) {
    bisection = Bisection(
        0.0, // acceptError: Stop if the search interval has
             // become smaller than this number
        1.0, // acceptFactor: ?
        parset.get<double>("bisection.requiredResidual"), true, // fastQuadratic
        0); // safety: acceptance factor for inexact minimization
  }

  /** \brief Constructs and returns an iterate object */
  IterateObject getIterateObject() {
    return IterateObject(parset, bisection, problem);
  }

private:
  // problem data
  MyConvexProblemType &problem;

  // commonly used minimization stuff
  Bisection bisection;

  Dune::ParameterTree const &parset;
};

/** \brief Solves one local system using a scalar Gauss-Seidel method */
template <class MyConvexProblemTypeTEMPLATE>
class MyBlockProblem<MyConvexProblemTypeTEMPLATE>::IterateObject {
  friend class MyBlockProblem;

protected:
  /** \brief Constructor, protected so only friends can instantiate it
   * \param bisection The class used to do a scalar bisection
   * \param problem The problem including quadratic part and nonlinear part
   */
  IterateObject(Dune::ParameterTree const &parset, Bisection const &bisection,
                MyConvexProblemType &problem)
      : parset(parset), problem(problem), bisection(bisection) {}

public:
  /** \brief Set the current iterate */
  void setIterate(VectorType &u) {
    this->u = u;
    return;
  }

  /** \brief Update the i-th block of the current iterate */
  void updateIterate(LocalVectorType const &ui, int i) {
    u[i] = ui;
    return;
  }

  /** \brief Minimise a local problem using a modified gradient method
   * \param[out] ui The solution
   * \param m Block number
   * \param ignore Set of degrees of freedom to leave untouched
   */
  void solveLocalProblem(
      LocalVectorType &ui, int m,
      const typename Dune::BitSetVector<block_size>::const_reference ignore) {
    {
      int ignore_component =
          block_size; // Special value that indicates nothing should be ignored
      switch (ignore.count()) {
        case 0: // Full problem
          break;
        case 1: // 1 Dimension is fixed
          assert(
              ignore[1]); // Only the second component is allowed to stay fixed
          ignore_component = 1;
          break;
        case block_size: // Ignore the whole node
          return;
        default:
          assert(false);
      }

      LocalMatrixType const *localA = NULL;
      LocalVectorType localb(problem.f[m]);

      typename MatrixType::row_type::ConstIterator it;
      typename MatrixType::row_type::ConstIterator end = problem.A[m].end();
      for (it = problem.A[m].begin(); it != end; ++it) {
        int const j = it.index();
        if (j == m)
          localA = &(*it); // localA = A[m][m]
        else
          it->mmv(u[j], localb); // localb -= A[m][j] * u[j]
      }
      assert(localA != NULL);

      auto phi = problem.phi.restriction(m);
      Dune::SampleFunctional<block_size> localJ(*localA, localb, phi,
                                                ignore_component);

      LocalVectorType correction;
      Dune::minimise(localJ, ui, parset.get<size_t>("localsolver.steps"),
                     bisection);
    }
  }

private:
  Dune::ParameterTree const &parset;

  // problem data
  MyConvexProblemType &problem;

  // commonly used minimization stuff
  Bisection bisection;

  // state data for smoothing procedure used by:
  // setIterate, updateIterate, solveLocalProblem
  VectorType u;
};

#endif
