#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_IPOPT
#undef HAVE_IPOPT
#endif

#include <exception>
#include <iostream>

#include "boost/format.hpp"

#include <dune/common/bitsetvector.hh>
#include <dune/common/exceptions.hh>
#include <dune/common/fmatrix.hh>
#include <dune/common/fvector.hh>
#include <dune/common/parametertree.hh>
#include <dune/common/parametertreeparser.hh>
#include <dune/common/shared_ptr.hh>
#include <dune/common/stdstreams.hh>
#include <dune/grid/common/mcmgmapper.hh>
#include <dune/grid/io/file/vtk/vtkwriter.hh>
#include <dune/grid/yaspgrid.hh>
#include <dune/istl/bcrsmatrix.hh>
#include <dune/istl/bvector.hh>

#include <dune/fufem/assemblers/boundaryfunctionalassembler.hh>
#include <dune/fufem/assemblers/functionalassembler.hh>
#include <dune/fufem/assemblers/localassemblers/neumannboundaryassembler.hh>
#include <dune/fufem/assemblers/localassemblers/stvenantkirchhoffassembler.hh>
#include <dune/fufem/assemblers/localassemblers/vonmisesstressassembler.hh>
#include <dune/fufem/assemblers/operatorassembler.hh>
#include <dune/fufem/boundarypatch.hh>
#include <dune/fufem/functions/basisgridfunction.hh>
#include <dune/fufem/functions/constantfunction.hh>
#include <dune/fufem/functions/vtkbasisgridfunction.hh>
#include <dune/fufem/functionspacebases/p0basis.hh>
#include <dune/fufem/functionspacebases/p1nodalbasis.hh>
#include <dune/solvers/common/numproc.hh> // Solver::FULL
#include <dune/solvers/iterationsteps/blockgsstep.hh>
#include <dune/solvers/iterationsteps/truncatedblockgsstep.hh>
#include <dune/solvers/norms/energynorm.hh>
#include <dune/solvers/solvers/loopsolver.hh>
#include <dune/tnnmg/iterationsteps/genericnonlineargs.hh>

#include <dune/tectonic/globallaursennonlinearity.hh>
#include <dune/tectonic/globalruinanonlinearity.hh>
#include <dune/tectonic/myconvexproblem.hh>
#include <dune/tectonic/myblockproblem.hh>

int const dim = 2;

template <class GridView>
void setup_boundary(GridView const &gridView,
                    Dune::FieldVector<double, dim> const &end_points,
                    Dune::BitSetVector<dim> &ignoreNodes,
                    Dune::BitSetVector<1> &neumannNodes,
                    Dune::BitSetVector<1> &frictionalNodes) {
  typedef typename GridView::template Codim<dim>::Iterator VertexLeafIterator;

  typedef Dune::MultipleCodimMultipleGeomTypeMapper<
      GridView, Dune::MCMGVertexLayout> VertexMapper;
  VertexMapper const myVertexMapper(gridView);

  size_t dirichlet_nodes = 0;
  size_t neumann_nodes = 0;
  size_t frictional_nodes = 0;
  for (VertexLeafIterator it = gridView.template begin<dim>();
       it != gridView.template end<dim>(); ++it) {
    Dune::GeometryType const gt = it->type();
    assert(it->geometry().corners() == 1);
    Dune::FieldVector<double, dim> const coordinates = it->geometry().corner(0);
    if (coordinates[1] == end_points[0]) {
      ++dirichlet_nodes;
      size_t const id = myVertexMapper.map(*it);
      ignoreNodes[id] = true;
    } else if (coordinates[1] == 0) {
      ++frictional_nodes;
      size_t const id = myVertexMapper.map(*it);
      frictionalNodes[id] = true;
      ignoreNodes[id][1] = true; // Zero displacement in direction y
    } else if (coordinates[0] == 0 || coordinates[0] == end_points[0]) {
      ++neumann_nodes;
      size_t const id = myVertexMapper.map(*it);
      neumannNodes[id] = true;
    }
  }
  std::cout << "Number of Neumann nodes: " << neumann_nodes << std::endl;
  std::cout << "Number of Dirichlet nodes: " << dirichlet_nodes << std::endl;
  std::cout << "Number of frictional nodes: " << dirichlet_nodes << std::endl;
}

// Assembles Neumann boundary term in f
template <class GridType, class GridView, class LocalVectorType, class FEBasis>
void assemble_neumann(GridView const &gridView, FEBasis const &feBasis,
                      Dune::BitSetVector<1> const &neumannNodes,
                      Dune::BlockVector<LocalVectorType> &f,
                      int run) { // constant sample function on neumann boundary
  BoundaryPatch<GridView> neumannBoundary(gridView, neumannNodes);
  LocalVectorType SampleVector(0);
  // FIXME: random values (time-dependent)
  SampleVector[0] =
      run / 10.0; // If we leave this fixed, we get into trouble for some reason
  SampleVector[1] = 0;
  ConstantFunction<LocalVectorType, LocalVectorType> fNeumann(SampleVector);
  NeumannBoundaryAssembler<GridType, LocalVectorType> neumannBoundaryAssembler(
      fNeumann);

  BoundaryFunctionalAssembler<FEBasis> boundaryFunctionalAssembler(
      feBasis, neumannBoundary);
  boundaryFunctionalAssembler.assemble(
      neumannBoundaryAssembler, f,
      true); // whether to resize the output vector and zero all of its entries
}

// Assembles constant 1-function on frictional boundary in nodalIntegrals
template <class GridType, class GridView, class LocalVectorType, class FEBasis>
void assemble_frictional(
    GridView const &gridView, FEBasis const &feBasis,
    Dune::BitSetVector<1> const &frictionalNodes,
    Dune::BlockVector<Dune::FieldVector<double, 1>> &nodalIntegrals) {
  BoundaryPatch<GridView> frictionalBoundary(gridView, frictionalNodes);
  ConstantFunction<LocalVectorType, Dune::FieldVector<double, 1>>
  constantOneFunction(1);
  NeumannBoundaryAssembler<GridType, Dune::FieldVector<double, 1>>
  frictionalBoundaryAssembler(constantOneFunction);

  BoundaryFunctionalAssembler<FEBasis> boundaryFunctionalAssembler(
      feBasis, frictionalBoundary);
  boundaryFunctionalAssembler.assemble(
      frictionalBoundaryAssembler, nodalIntegrals,
      true); // whether to resize the output vector and zero all of its entries
}

void assemble_nonlinearity(
    int size, Dune::ParameterTree const &parset,
    Dune::shared_ptr<Dune::GlobalNonlinearity<dim> const> &myGlobalNonlinearity,
    Dune::shared_ptr<Dune::BlockVector<Dune::FieldVector<double, 1>>>
        nodalIntegrals) {
  // {{{ Assemble terms for the nonlinearity
  auto mu =
      Dune::make_shared<Dune::BlockVector<Dune::FieldVector<double, 1>>>(size);
  *mu = parset.get<double>("boundary.friction.mu");

  auto normalStress =
      Dune::make_shared<Dune::BlockVector<Dune::FieldVector<double, 1>>>(size);
  *normalStress = parset.get<double>("boundary.friction.normalstress");

  std::string const friction_model =
      parset.get<std::string>("boundary.friction.model");
  if (friction_model == std::string("Ruina")) {
    auto a = Dune::make_shared<Dune::BlockVector<Dune::FieldVector<double, 1>>>(
        size);
    *a = parset.get<double>("boundary.friction.ruina.a");

    auto eta =
        Dune::make_shared<Dune::BlockVector<Dune::FieldVector<double, 1>>>(
            size);
    *eta = parset.get<double>("boundary.friction.eta");

    myGlobalNonlinearity =
        Dune::make_shared<Dune::GlobalRuinaNonlinearity<dim> const>(
            nodalIntegrals, a, mu, eta, normalStress);
  } else if (friction_model == std::string("Laursen")) {
    myGlobalNonlinearity = Dune::make_shared<
        Dune::GlobalLaursenNonlinearity<dim, Dune::LinearFunction> const>(
        mu, normalStress, nodalIntegrals);
  } else {
    assert(false);
  }
}

int main(int argc, char *argv[]) {
  try {
    Dune::ParameterTree parset;
    Dune::ParameterTreeParser::readINITree("../../src/one-body-sample.parset",
                                           parset);
    Dune::ParameterTreeParser::readOptions(argc, argv, parset);

    typedef Dune::FieldVector<double, dim> SmallVector;
    typedef Dune::FieldMatrix<double, dim, dim> SmallMatrix;
    typedef Dune::BCRSMatrix<SmallMatrix> OperatorType;
    typedef Dune::BlockVector<SmallVector> VectorType;
    typedef Dune::BlockVector<Dune::FieldVector<double, 1>> CellVectorType;

    auto const E = parset.get<double>("body.E");
    auto const nu = parset.get<double>("body.nu");
    auto const solver_maxIterations =
        parset.get<size_t>("solver.maxiterations");
    auto const solver_tolerance = parset.get<double>("solver.tolerance");

    auto const levels = parset.get<int>("grid.refinements") + 1;

    auto const verbose = parset.get<bool>("verbose");
    Solver::VerbosityMode verbosity = verbose ? Solver::FULL : Solver::QUIET;

    // {{{ Set up grid
    typedef Dune::YaspGrid<dim> GridType;
    Dune::FieldVector<double, dim> const end_points(
        1); // nth dimension (zero-indexed) goes from 0 to end_points[n]
    GridType grid(
        end_points,
        Dune::FieldVector<int, dim>(2), // number of elements in each direction
        Dune::FieldVector<bool, dim>(false), // non-periodic in each direction
        0);                                  // zero overlap (whatever that is)
    grid.globalRefine(levels - 1);

    typedef GridType::LeafGridView GridView;
    GridView const leafView = grid.leafView();
    // }}}

    // Set up bases
    typedef P0Basis<GridView, double> P0Basis;
    typedef P1NodalBasis<GridView, double> P1Basis;
    P0Basis const p0Basis(leafView);
    P1Basis const p1Basis(leafView);

    // Assemble elastic force on the body
    StVenantKirchhoffAssembler<GridType, P1Basis::LocalFiniteElement,
                               P1Basis::LocalFiniteElement> const
    localStiffness(E, nu);
    OperatorType stiffnessMatrix;
    OperatorAssembler<P1Basis, P1Basis>(p1Basis, p1Basis)
        .assemble(localStiffness, stiffnessMatrix);
    EnergyNorm<OperatorType, VectorType> energyNorm(stiffnessMatrix);

    // Set up the boundary
    Dune::BitSetVector<dim> ignoreNodes(grid.size(grid.maxLevel(), dim), false);
    Dune::BitSetVector<1> neumannNodes(grid.size(grid.maxLevel(), dim), false);
    Dune::BitSetVector<1> frictionalNodes(grid.size(grid.maxLevel(), dim),
                                          false);
    setup_boundary(leafView, end_points, ignoreNodes, neumannNodes,
                   frictionalNodes);

    typedef MyConvexProblem<OperatorType, VectorType> MyConvexProblemType;
    typedef MyBlockProblem<MyConvexProblemType> MyBlockProblemType;
    GenericNonlinearGS<MyBlockProblemType> nonlinearGSStep;
    nonlinearGSStep.ignoreNodes_ = &ignoreNodes;

    VectorType u1(grid.size(grid.maxLevel(), dim));
    u1 = 0;
    VectorType u2 = u1;
    VectorType u3 = u1;

    VectorType u1_diff_new(grid.size(grid.maxLevel(), dim));
    u1_diff_new = 0;
    VectorType u2_diff_new = u1_diff_new;
    VectorType u3_diff_new = u1_diff_new;

    CellVectorType vonMisesStress;

    VectorType b1;
    VectorType b2;
    VectorType b3;

    auto nodalIntegrals =
        Dune::make_shared<Dune::BlockVector<Dune::FieldVector<double, 1>>>();
    assemble_frictional<GridType, GridView, SmallVector, P1Basis>(
        leafView, p1Basis, frictionalNodes, *nodalIntegrals);

    Dune::shared_ptr<Dune::GlobalNonlinearity<dim> const> myGlobalNonlinearity;
    assemble_nonlinearity(grid.size(grid.maxLevel(), dim), parset,
                          myGlobalNonlinearity, nodalIntegrals);

    for (size_t run = 1; run <= parset.get<size_t>("timesteps"); ++run) {
      std::cout << ".";
      std::cout.flush();

      // b = neumann
      assemble_neumann<GridType, GridView, SmallVector, P1Basis>(
          leafView, p1Basis, neumannNodes, b1, run);
      b2 = b1;
      b3 = b1;

      // b -= linear update
      stiffnessMatrix.mmv(u1, b1);
      stiffnessMatrix.mmv(u2, b2);
      stiffnessMatrix.mmv(u3, b3);

      if (parset.get<bool>("useNonlinearGS")) {
        MyConvexProblemType myConvexProblem(
            stiffnessMatrix, *myGlobalNonlinearity, b1, u1_diff_new);
        MyBlockProblemType myBlockProblem(parset, myConvexProblem);
        nonlinearGSStep.setProblem(u1_diff_new, myBlockProblem);

        LoopSolver<VectorType> solver(&nonlinearGSStep, solver_maxIterations,
                                      solver_tolerance, &energyNorm, verbosity);
        solver.solve();
      }

      u1 += u1_diff_new;

      { // Compute von Mises stress and write everything to a file
        auto *displacement =
            new BasisGridFunction<P1Basis, VectorType>(p1Basis, u1);
        VonMisesStressAssembler<GridType> localStressAssembler(E, nu,
                                                               displacement);
        FunctionalAssembler<P0Basis>(p0Basis)
            .assemble(localStressAssembler, vonMisesStress, true);

        Dune::VTKWriter<GridView> writer(leafView);
        std::string filename((boost::format("obs%d") % run).str());

        auto displacement_ptr =
            Dune::make_shared<VTKBasisGridFunction<P1Basis, VectorType>>(
                p1Basis, u1, "displacement");
        auto vonmises_ptr =
            Dune::make_shared<VTKBasisGridFunction<P0Basis, CellVectorType>>(
                p0Basis, vonMisesStress, "stress");
        writer.addVertexData(displacement_ptr);
        writer.addCellData(vonmises_ptr);
        writer.write(filename.c_str());
      }

      if (parset.get<bool>("useGS")) {
        BlockGSStep<OperatorType, VectorType> blockGSStep(stiffnessMatrix,
                                                          u2_diff_new, b2);
        blockGSStep.ignoreNodes_ = &ignoreNodes;

        LoopSolver<VectorType> solver(&blockGSStep, solver_maxIterations,
                                      solver_tolerance, &energyNorm, verbosity);
        solver.solve();
      }

      u2 += u2_diff_new;

      // Use a linear solver for comparison; should return roughly the
      // same results if phi vanishes (e.g. because the normalstress is zero)
      if (parset.get<bool>("useTruncatedGS")) {
        TruncatedBlockGSStep<OperatorType, VectorType> blockGSStep(
            stiffnessMatrix, u3_diff_new, b3);
        blockGSStep.ignoreNodes_ = &ignoreNodes;

        LoopSolver<VectorType> solver(&blockGSStep, solver_maxIterations,
                                      solver_tolerance, &energyNorm, verbosity);
        solver.solve();
      }

      u3 += u3_diff_new;
    }
    std::cout << std::endl;

    VectorType diff2 = u2;
    diff2 -= u1;
    std::cout << "sup |u1 - u2| = " << diff2.infinity_norm() << ", "
              << "|u1 - u2| = " << diff2.two_norm() << std::endl;

    VectorType diff3 = u3;
    diff3 -= u1;
    std::cout << "sup |u1 - u3| = " << diff3.infinity_norm() << ", "
              << "|u1 - u3| = " << diff3.two_norm() << std::endl;

    // Print displacement on frictional boundary
    for (size_t i = 0; i < frictionalNodes.size(); ++i)
      if (frictionalNodes[i][0])
        std::cout << boost::format("u1[%02d] = %+3e, %|40t|u2[%02d] = %+3e "
                                   "%|80t|u3[%02d] = %+3e") %
                         i % u1[i] % i % u2[i] % i % u3[i] << std::endl;
  }
  catch (Dune::Exception &e) {
    Dune::derr << "Dune reported error: " << e << std::endl;
  }
  catch (std::exception &e) {
    std::cerr << "Standard exception: " << e.what() << std::endl;
  }
}
